/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include <math.h>

#include "kftpqueue.h"
#include "kftpbookmarks.h"
#include "kftpgrabber.h"
#include "kftpconfiguration.h"
#include "kftpsystemtray.h"
#include "kftpqueueprocessor.h"
#include "kftpsession.h"

#include <kmessagebox.h>
#include <klocale.h>
#include <kio/renamedlg.h>
#include <kdiskfreesp.h>
#include <kfileitem.h>

#include <qapplication.h>
#include <qregexp.h>
#include <qdir.h>
#include <qobjectlist.h>

KFTPQueueManager *FTPQueueManager;

namespace KFTPQueue {

QueueScannerThread::QueueScannerThread(KFTPSession *session, Transfer *item)
  : QThread(), m_session(session), m_item(item)
{
}

void QueueScannerThread::run()
{
  // First scan the folder
  scanFolder(m_item);

  // We are done, post event to notify the GUI
  QueueScannerThreadEvent *e = new QueueScannerThreadEvent(m_item);
  qApp->postEvent(m_session, e);
}

void QueueScannerThread::scanFolder(Transfer *parent)
{
  QDir fs(parent->getSourceUrl().path());
  fs.setFilter(QDir::Readable | QDir::Hidden | QDir::All);

  const QFileInfoList *p_list = fs.entryInfoList();
  QFileInfoListIterator i(*p_list);
  QFileInfo *file;

  while ((file = i.current()) != 0) {
    ++i;

    if (file->fileName() == "." || file->fileName() == "..") continue;
    if (FTPConfig().m_skipEmptyFiles && file->size() == 0) continue;

    // Spawn transfer
    KURL destUrl = parent->getDestUrl();
    destUrl.addPath(file->fileName());

    // This is needed, since QFileInfo works with uint for the filesize
    filesize_t realSize = KFileItem(KFileItem::Unknown, KFileItem::Unknown, file->absFilePath()).size();

    if (file->isDir()) {
      // Directory
      KFTPQueue::TransferDir *transfer = new KFTPQueue::TransferDir(parent);
      transfer->setSourceUrl(KURL(file->absFilePath()));
      transfer->setDestUrl(destUrl);
      transfer->addSize(realSize);
      transfer->setTransferType(parent->getTransferType());
      transfer->setId(FTPQueueManager->p_lastQID++);

      qApp->lock();
      emit FTPQueueManager->newTransfer(transfer);
      qApp->unlock();

      // Call this function in recursion
      scanFolder(transfer);

      if (FTPConfig().m_skipEmptyDirs && !transfer->children()) {
        qApp->lock();
        FTPQueueManager->removeTransfer(transfer);
        qApp->unlock();
      }
    } else {
      // Spawn a new child transfer
      if (FTPConfig().m_enableSkipList && FTPConfig().m_skipNoQueue && FTPConfig().isOnSkipList(file->absFilePath()))
        continue;

      if (FTPConfig().m_skipEmptyFiles && realSize == 0)
        continue;

      KFTPQueue::TransferFile *transfer = new KFTPQueue::TransferFile(parent);
      transfer->setSourceUrl(KURL(file->absFilePath()));
      transfer->setDestUrl(destUrl);
      transfer->addSize(realSize);
      transfer->setTransferType(parent->getTransferType());
      transfer->setId(FTPQueueManager->p_lastQID++);

      qApp->lock();
      emit FTPQueueManager->newTransfer(transfer);
      qApp->unlock();
    }
  }
}

}

using namespace KFTPQueue;

KFTPQueueManager::KFTPQueueManager(QObject *parent)
  : QObject(parent)
{
  p_queueTransfers.setAutoDelete(false);
  p_lastQID = 0;
  m_curDownSpeed = 0;
  m_curUpSpeed = 0;
  p_emitUpdate = true;

  // Create the queue processor object
  m_queueProc = new KFTPQueueProcessor(this);

  connect(m_queueProc, SIGNAL(queueComplete()), this, SLOT(slotQueueProcessingComplete()));

  // Create the queue converter object
  m_converter = new KFTPQueueConverter(this);
}

void KFTPQueueManager::insertTransfer(Transfer *transfer)
{
  transfer->setId(p_lastQID++);
  p_queueTransfers.append(transfer);

  /* Connect to some signals */
  connect(transfer, SIGNAL(transferComplete(long)), this, SLOT(slotTransferComplete(long)));
  connect(transfer, SIGNAL(transferStart(long)), this, SLOT(slotTransferStart(long)));
  connect(transfer, SIGNAL(transferAbort(long)), this, SLOT(slotTransferAbort(long)));

  emit newTransfer(transfer);
  if (p_emitUpdate)
    emit queueUpdate();
}

KFTPQueueTransfers *KFTPQueueManager::getQueueTransferList()
{
  return &p_queueTransfers;
}

void KFTPQueueManager::removeTransfer(Transfer *transfer)
{
  if (!transfer)
    return;
    
  transfer->abort();
  long id = transfer->getId();

  // Signal destruction
  transfer->faceDestruction();

  if (!transfer->isChild()) {
    p_queueTransfers.remove(transfer);
  }

  /* Actually delete the transfer */
  delete transfer;

  emit transferRemoved(id);
  if (p_emitUpdate)
    emit queueUpdate();
}

void KFTPQueueManager::removeTransfer(long id)
{
  // Just remove the transfer we find via id
  removeTransfer(findTransfer(id));
}

void KFTPQueueManager::clearQueue()
{
  while (p_queueTransfers.count() > 0) {
    Transfer *i;
    for (i = p_queueTransfers.first(); i; i = p_queueTransfers.next()) {
      removeTransfer(i);
    }
  }

  if (p_emitUpdate)
    emit queueUpdate();
}

Transfer *KFTPQueueManager::findTransfer(const QString &sourcePath)
{
  Transfer *i;
  for (i = p_queueTransfers.first(); i; i = p_queueTransfers.next()) {
    if (i->getSourceUrl().path() == sourcePath)
      return i;
  }

  return NULL;
}

Transfer *KFTPQueueManager::findTransfer(long id)
{
  Transfer *i;
  for (i = p_queueTransfers.first(); i; i = p_queueTransfers.next()) {
    if (i->getId() == id)
      return i;

    if (i->isDir()) {
      // This transfer has children
      Transfer *child = static_cast<TransferDir*>(i)->findChildTransfer(id);

      if (child)
        return child;
    }
  }

  return NULL;
}

void KFTPQueueManager::moveTransferUp(long id)
{
  Transfer *transfer = findTransfer(id);
  if (transfer->isChild()) {
    static_cast<TransferDir*>(transfer->getTransferParent())->moveChildUp(transfer);
  } else {
    if (p_queueTransfers.findRef(transfer) != -1) {
      if (p_queueTransfers.prev()) {
        int prevPos = p_queueTransfers.at();
        p_queueTransfers.remove(transfer);
        p_queueTransfers.insert(prevPos, transfer);
      }
    }
  }

  if (p_emitUpdate)
    emit queueUpdate();
}

void KFTPQueueManager::moveTransferDown(long id)
{
  Transfer *transfer = findTransfer(id);
  if (transfer->isChild()) {
    static_cast<TransferDir*>(transfer->getTransferParent())->moveChildDown(transfer);
  } else {
    if (p_queueTransfers.findRef(transfer) != -1) {
      if (p_queueTransfers.next()) {
        int nextPos = p_queueTransfers.at();
        p_queueTransfers.remove(transfer);
        p_queueTransfers.insert(nextPos, transfer);
      }
    }
  }

  if (p_emitUpdate)
    emit queueUpdate();
}

bool KFTPQueueManager::canBeMovedUp(long id)
{
  Transfer *transfer = findTransfer(id);
  
  if (transfer->isChild() && (!static_cast<TransferDir*>(transfer->getTransferParent())->canMoveChildUp(transfer))) return false;
  if (transfer->isRunning()) return false;
  if (p_queueTransfers.first() == transfer) return false;

  return true;
}

bool KFTPQueueManager::canBeMovedDown(long id)
{
  Transfer *transfer = findTransfer(id);
  
  if (transfer->isChild() && (!static_cast<TransferDir*>(transfer->getTransferParent())->canMoveChildDown(transfer))) return false;
  if (transfer->isRunning()) return false;
  if (p_queueTransfers.last() == transfer) return false;

  return true;
}

filesize_t KFTPQueueManager::getDownloadSpeed()
{
  return m_curDownSpeed;
}

filesize_t KFTPQueueManager::getUploadSpeed()
{
  return m_curUpSpeed;
}

int KFTPQueueManager::getTransferPercentage()
{
  /* Go trough all transfers and calculate current percentage of completion */
  if (p_queueTransfers.count() == 0)
    return 0;

  Transfer *i;
  int p_percentage = 0;

  for (i = p_queueTransfers.first(); i; i = p_queueTransfers.next()) {
    if (i->getSize() > 0) {
      p_percentage += (i->getCompleted()*100)/i->getSize();
    } else {
      // Unknown file size - ignore it
    }
  }

  return (int) rint(p_percentage/p_queueTransfers.count());
}

int KFTPQueueManager::getNumRunning(bool onlyDirs)
{
  // Return the number of running transfers
  if (p_queueTransfers.count() == 0)
    return 0;

  Transfer *i;
  int p_running = 0;

  for (i = p_queueTransfers.first(); i; i = p_queueTransfers.next()) {
    if (i->isRunning() && (!onlyDirs || i->isDir()))
      p_running++;
  }

  return p_running;
}

void KFTPQueueManager::doEmitUpdate()
{
  /* Check if any objects are marked for deletion and if not, get the dl/ul speed */
  m_curDownSpeed = 0;
  m_curUpSpeed = 0;

  Transfer *i;
  for (i = p_queueTransfers.first(); i; i = p_queueTransfers.next()) {
    if (i->isDir())
      static_cast<TransferDir*>(i)->removeMarkedChildren();

    if (i->isDeleteMarked()) {
      /* Transfer asked to be deleted, let it be so! */
      removeTransfer(i);
    } else {
      if (i->getTransferType() == Upload)
        m_curUpSpeed += i->getSpeed();
      else if (i->getTransferType() == Download)
        m_curDownSpeed += i->getSpeed();
    }
  }

  // Emit global update to all GUI objects
  emit queueUpdate();
}

void KFTPQueueManager::setEmitUpdate(bool value)
{
  p_emitUpdate = value;
}

Transfer *KFTPQueueManager::spawnTransfer(KURL srcURL, KURL dstURL, filesize_t size, bool dir, TransferType t_type, bool ignoreSkip,
                                          bool insertToQueue, QObject *parent, bool noScan)
{
  if (!ignoreSkip && FTPConfig().m_enableSkipList && FTPConfig().m_skipNoQueue && FTPConfig().isOnSkipList(srcURL.path()))
    return 0L;

  // Should we lowercase the destination path ?
  if (FTPConfig().m_lowercaseDownload && t_type == Download) {
    dstURL.setPath(dstURL.directory() + "/" + dstURL.fileName().lower());
  }

  if (!parent)
    parent = this;

  Transfer *transfer = 0L;

  if (dir)
    transfer = new TransferDir(parent);
  else
    transfer = new TransferFile(parent);

  transfer->setSourceUrl(srcURL);
  transfer->setDestUrl(dstURL);
  transfer->addSize(size);
  transfer->setTransferType(t_type);

  if (insertToQueue) {
    insertTransfer(transfer);
  } else {
    transfer->setId(p_lastQID++);
    emit newTransfer(transfer);
  }

  if (dir && !noScan) {
    // This is a directory, we should scan the directory and add all files/dirs found
    // as parent of current object
    KFTPSession *session = FTPSessionManager->spawnRemoteSession(IGNORE_SIDE, srcURL);
    session->scanDirectory(transfer);
  }

  return transfer;
}

void KFTPQueueManager::insertTransfer(KURLDrag *drag)
{
  // Decode the drag
  KIO::MetaData p_meta;
  KURL::List p_urls;
  KURLDrag::decode(drag, p_urls, p_meta);

  // TODO make support for local drops - eg. from konqueror, where
  // we get no meta data, so we must get the file info ourselves and
  // reject remote urls (or show a dialog to ask the user if he
  // wants to connect to the remote site)

  // Now we should add transfers for all URLs
  Transfer *lastTransfer = 0L;
  KURL::List::iterator end(p_urls.end());

  for (KURL::List::iterator i(p_urls.begin()); i != end; ++i) {
    QString p_data = p_meta[(*i).htmlURL().local8Bit()];
    QChar p_type = p_data.at(0);
    filesize_t p_size = p_data.section(':', 1, 1).toULongLong();
    KURL p_src = (*i);
    KURL p_dst = KURL(p_meta["DestURL"]);
    p_dst.addPath(p_src.fileName());

    // Skip where both files are local
    if (p_src.isLocalFile() && p_dst.isLocalFile())
      continue;

    // Determine transfer type
    TransferType p_trans = p_src.isLocalFile() ? Upload : (p_dst.isLocalFile() ? Download : FXP);

    lastTransfer = spawnTransfer(p_src, p_dst, p_size, p_type == 'D', p_trans, true, true, 0L, true);
  }

  // Execute the transfer if set in configuration
  if (!FTPConfig().m_queueOnDND && p_urls.count() == 1 && lastTransfer)
    lastTransfer->delayedExecute(100);
}

int KFTPQueueManager::fileExistsDialog(KFTPQueue::TransferFile *transfer, const QString &local, FTPEntry remote, bool reversed)
{
  // reversed == false | DOWNLOAD
  // reversed == true  | UPLOAD
  QFileInfo localInfo(local);

  if (!remote.name.isEmpty() && localInfo.exists()) {
    filesize_t srcSize = reversed ? localInfo.size() : remote.size;
    filesize_t dstSize = reversed ? remote.size : localInfo.size();

    time_t srcTime = reversed ? remote.size : localInfo.created().toTime_t();
    time_t dstTime = reversed ? localInfo.created().toTime_t() : remote.size;

    FileExistsActions fa = reversed ? FTPConfig().m_fileExistsUpActions : FTPConfig().m_fileExistsDownActions;
    FEAction act = fa.getActionForSituation(srcSize, srcTime, dstSize, dstTime);

    switch (act) {
      case FE_SKIP_ACT: {
        if (transfer->isChild() && transfer->getTransferParent()->isRunning()) {
          transfer->m_deleteMe = true;
          transfer->addActualSize(-transfer->m_size);
        }

        transfer->resetTransfer();
        emit transfer->transferComplete(transfer->m_id);

        FTPQueueManager->doEmitUpdate();
        return -1;
      }
      case FE_OVERWRITE_ACT: {
        return KIO::R_OVERWRITE;
      }
      case FE_RESUME_ACT: {
        return KIO::R_RESUME;
      }
      case FE_RENAME_ACT: {
        // TODO implement rename action
      }
      case FE_USER_ACT: {
        QString newDestPath;
        KIO::RenameDlg_Result r = KIO::open_RenameDlg(
          i18n("File Exists"),
          transfer->getSourceUrl().prettyURL(),
          transfer->getDestUrl().prettyURL(),
          (KIO::RenameDlg_Mode) (KIO::M_OVERWRITE | KIO::M_RESUME),
          newDestPath,
          srcSize,
          dstSize,
          srcTime,
          dstTime
        );

        switch (r) {
          case KIO::R_RENAME: {
            KURL url = transfer->getDestUrl();
            url.setPath(newDestPath);
            transfer->setDestUrl(url);
            return 0;
          }
          case KIO::R_CANCEL: {
            if (transfer->isChild() && transfer->getTransferParent()->isRunning()) {
              transfer->m_deleteMe = true;
              transfer->addActualSize(-transfer->m_size);
            }

            transfer->resetTransfer();
            emit transfer->transferComplete(transfer->m_id);

            FTPQueueManager->doEmitUpdate();
            return -1;
          }
          default: return r;
        }
      }
    }
  }
  return true;
}

void KFTPQueueManager::sortTransfersByServer()
{
  // Sort all transfers, so that transfers on the same server/account will be grouped together
  // for further processing

  // TODO sorting algorythm should also sort by priority when this is implemented
  // FIXME we need a better sorting algorythm - actually current sorting algo. sux ;)
  // FIXME quick sort should be used ?
  int sortPos;
  KURL sortKey;

  for (unsigned int i = 0; i < p_queueTransfers.count(); ++i) {
    Transfer *transfer = p_queueTransfers.at(i);

    // Set sortkey and position
    switch (transfer->getTransferType()) {
      case Download: sortKey = transfer->getSourceUrl(); break;
      case Upload: sortKey = transfer->getDestUrl(); break;
      case FXP: sortKey = transfer->getSourceUrl(); break;
    }
    sortPos = i;

    // Group all that matches sortKey together
    for (unsigned int j = i + 1; j < p_queueTransfers.count(); ++j) {
      Transfer *sTransfer = p_queueTransfers.at(j);

      KURL tmpUrl;
      switch (transfer->getTransferType()) {
        case Download: tmpUrl = transfer->getSourceUrl(); break;
        case Upload: tmpUrl = transfer->getDestUrl(); break;
        case FXP: tmpUrl = transfer->getSourceUrl(); break;
      }
      tmpUrl.setPath(sortKey.path());

      if (tmpUrl == sortKey) {
        int n = j;
        while (n != sortPos + 1) {
          moveTransferUp(sTransfer->getId());
          n--;
        }

        sortPos = n;
      }
    }
  }

  // Now everything should be sorted
  emit queueUpdate();
}

void KFTPQueueManager::startQueueTransfer()
{
  // First, sort the transfers
  sortTransfersByServer();

  // Now, go trough all queued files and execute them - try to do as little server connects
  // as possible
  m_queueProc->startProcessing();
}

void KFTPQueueManager::abortQueueTransfer()
{
  // Stop further queue processing
  m_queueProc->stopProcessing();

  emit queueUpdate();
}

void KFTPQueueManager::slotQueueProcessingComplete()
{
  // Queue processing is now complete
  if (FTPConfig().m_showBalloons) {
    QString p_finishedStr = i18n("All queued transfers have been completed!");
    s_sysTray->showBalloon(p_finishedStr);
  }

  emit queueUpdate();
}

void KFTPQueueManager::slotTransferComplete(long id)
{
  emit qsaTransferComplete(id);
}

void KFTPQueueManager::slotTransferStart(long id)
{
  emit qsaTransferStart(id);
}

void KFTPQueueManager::slotTransferAbort(long id)
{
  emit qsaTransferAbort(id);
}



#include "kftpqueue.moc"
