/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2004 by the KFTPGrabber developers
 * Copyright (C) 2004 Jernej Kos <kostko@jweb-network.net>
 * Copyright (C) 2004 Markus Brueffer <markus@brueffer.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpapi.h"
#include "kftpbookmarkeditorlistview.h"
#include "kftpbookmarks.h"
#include "kftpsession.h"
#include "misc.h"

#include <qdragobject.h>

#include <kaction.h>
#include <kinputdialog.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kdebug.h>

using namespace KFTPGrabberBase;

KFTPBookmarkEditorItem::KFTPBookmarkEditorItem(QListView *parent, const QString &t1)
  : KListViewItem(parent, t1)
{
}

KFTPBookmarkEditorItem::KFTPBookmarkEditorItem(QListViewItem *parent, const QString &t1)
  : KListViewItem(parent, t1)
{
}

int KFTPBookmarkEditorItem::compare(QListViewItem *i, int col, bool) const
{
  if (m_type != ((KFTPBookmarkEditorItem*)i)->m_type) {
    if (m_type == 0)
      return -1; // category
    else
      return 1; // server
  }

  return KListViewItem::compare(i, col, false);
}

KFTPBookmarkEditorListView::KFTPBookmarkEditorListView(KFTPBookmarks::Manager *bookmarks, QWidget *parent, const char *name)
 : KFTPListView(parent, name),
   m_autoUpdate(false),
   m_connectBookmark(false),
   m_editMenuItem(false),
   m_bookmarks(bookmarks),
   m_activeSite(0),
   m_activeCategory(0),
   m_activeItem(0)
{
  // Set some stuff
  setMinimumWidth(150);
  setFullWidth(true);
  addColumn(i18n("Bookmarks"));
  setRootIsDecorated(true);
  setEmptyListText(i18n("No bookmarks."));
  setItemsRenameable(false);

  // Drag & drop
  setDragEnabled(true);
  setAcceptDrops(true);
  setDropVisualizer(false);

  // Init auto open timer
  m_openTimer = new QTimer(this);

  connect(this, SIGNAL(dropped(QDropEvent*, QListViewItem*)), this, SLOT(slotDropped(QDropEvent*, QListViewItem*)));
  connect(this, SIGNAL(clicked(QListViewItem*)), this, SLOT(slotClicked(QListViewItem*)));
  connect(this, SIGNAL(doubleClicked(QListViewItem*)), this, SLOT(slotDoubleClicked(QListViewItem*)));
  connect(this, SIGNAL(contextMenuRequested(QListViewItem*, const QPoint&, int)), this, SLOT(slotContextMenu(QListViewItem*, const QPoint&, int)));

  /* Init the actions */
  m_newAction = new KAction(i18n("&New..."), "filenew", KShortcut(), this, SLOT(slotNewAction()), actionCollection(), "bookmark_new");
  m_renameAction = new KAction(i18n("&Rename"), KShortcut(), this, SLOT(slotRenameAction()), actionCollection(), "bookmark_rename");
  m_deleteAction = new KAction(i18n("&Delete"), "editdelete", KShortcut(), this, SLOT(slotDeleteAction()), actionCollection(), "bookmark_delete");
  m_subCatAction = new KAction(i18n("&Create Subcategory..."), "folder_new", KShortcut(), this, SLOT(slotSubCatAction()), actionCollection(), "bookmark_subcat");
  m_copyAction = new KAction(i18n("&Duplicate"), KShortcut(), this, SLOT(slotDuplicateAction()), actionCollection(), "bookmark_duplicate");
}

KActionCollection *KFTPBookmarkEditorListView::actionCollection()
{
  return KFTPAPI::getInstance()->mainWindow()->actionCollection();
}

void KFTPBookmarkEditorListView::setAutoUpdate(bool value)
{
  m_autoUpdate = value;

  if (value)
    connect(m_bookmarks, SIGNAL(update()), this, SLOT(slotAutoUpdate()));

  slotAutoUpdate();
}

void KFTPBookmarkEditorListView::setConnectBookmark(bool value)
{
  m_connectBookmark = value;

  if (value)
    connect(this, SIGNAL(executed(QListViewItem*)), this, SLOT(slotBookmarkExecuted(QListViewItem*)));
}

void KFTPBookmarkEditorListView::setEditMenuItem(bool value)
{
  m_editMenuItem = value;
}

void KFTPBookmarkEditorListView::slotBookmarkExecuted(QListViewItem *item)
{
  if (!item || !m_connectBookmark || static_cast<KFTPBookmarkEditorItem*>(item)->m_type == BT_CATEGORY)
    return;

  KFTPBookmarks::Site *site = static_cast<KFTPBookmarkEditorItem*>(item)->m_site;
  KURL url = site->getUrl();

  // Check something
  if (!url.hasUser()) {
    KMessageBox::sorry(0, i18n("No username specified for '%1'!").arg(url.host()), i18n("Error Connecting"));
    return;
  }

  // Just spawn a new session
  KFTPSession *session = FTPSessionManager->spawnRemoteSession(IGNORE_SIDE, url, site);
  m_bookmarks->setupClient(site, session->getClient());
}

void KFTPBookmarkEditorListView::slotAutoUpdate()
{
  if (m_autoUpdate) {
    // Update the bookmark list
    fillBookmarkData();
  }
}

QDragObject *KFTPBookmarkEditorListView::dragObject()
{
  // Determine hotspot and pixmap
  QPoint hotspot;
  QPixmap pixmap = *selectedItem()->pixmap(0);
  hotspot.setX(pixmap.width() / 2);
  hotspot.setY(pixmap.height() / 2);

  QString id = static_cast<KFTPBookmarkEditorItem*>(selectedItem())->m_site->id();
  m_dragObject = new QStoredDrag("application/x-qlistviewitem", this);
  QByteArray data;
  QDataStream arg(data, IO_WriteOnly);
  arg << id;

  // Start the drag
  static_cast<QStoredDrag*>(m_dragObject)->setEncodedData(data);
  m_dragObject->setPixmap(pixmap, hotspot);

  return m_dragObject;
}

bool KFTPBookmarkEditorListView::acceptDrag(QDropEvent *e) const
{
  // If it is a local drag, accept it
  return e->source() == static_cast<QWidget*>(const_cast<KFTPBookmarkEditorListView*>(this));
}

void KFTPBookmarkEditorListView::contentsDragEnterEvent(QDragEnterEvent *e)
{
  if (!acceptDrag(e)) {
    e->ignore();
    return;
  }
  e->acceptAction();

  m_currentBeforeDropItem = selectedItem();
  QListViewItem *item = itemAt(contentsToViewport(e->pos()));

  if (item) {
    m_dropItem = item;
    m_openTimer->start(750);
  } else {
    m_dropItem = 0L;
  }
}

void KFTPBookmarkEditorListView::contentsDragMoveEvent(QDragMoveEvent *e)
{
  if (!acceptDrag(e)) {
    e->ignore();
    return;
  }
  e->acceptAction();

  QListViewItem *item = itemAt(contentsToViewport(e->pos()));
  if (item) {
    setSelected(item, true);
    if (item != m_dropItem) {
      m_openTimer->stop();
      m_dropItem = item;
      m_openTimer->start(750);
    }
  } else {
    if (selectedItem()) {
      setSelected(selectedItem(), false);
    }

    m_openTimer->stop();
    m_dropItem = 0L;
  }
}

void KFTPBookmarkEditorListView::contentsDragLeaveEvent(QDragLeaveEvent*)
{
  if (m_currentBeforeDropItem) {
    setSelected(m_currentBeforeDropItem, true);
    ensureItemVisible(m_currentBeforeDropItem);
  } else {
    setSelected(m_dropItem, false);
  }

  m_openTimer->stop();
  m_dropItem = 0L;
}

void KFTPBookmarkEditorListView::slotOpenTimer()
{
  m_openTimer->stop();
  if (m_dropItem && m_dropItem->isExpandable()) {
    // Expand the category
    m_dropItem->setOpen(true);
  }
}

void KFTPBookmarkEditorListView::slotDropped(QDropEvent *e, QListViewItem*)
{
  KFTPBookmarks::Site *parentSite = 0L;

  if (selectedItem()) {
    QListViewItem *newParent = 0L;

    // Get the new path
    if (static_cast<KFTPBookmarkEditorItem*>(selectedItem())->m_type == 0)
      newParent = selectedItem();
    else
      newParent = selectedItem()->parent();

    if (newParent)
      parentSite = static_cast<KFTPBookmarkEditorItem*>(newParent)->m_site;
    else
      parentSite = m_bookmarks->findCategory("root");
  }

  QString id;

  // Decode the data
  QDataStream arg(e->encodedData("application/x-qlistviewitem"), IO_ReadOnly);
  arg >> id;

  // Move the site
  KFTPBookmarks::Site *originalSite = m_bookmarks->findSite(id);
  parentSite->reparentSite(originalSite);

  emit bookmarkMoved();

  // Notify the bookmark manager
  m_bookmarks->emitUpdate();
}

void KFTPBookmarkEditorListView::fillBookmarkData()
{
  // Fill the tree with data
  m_bookmarks->guiPopulateBookmarksTree(this);
}

void KFTPBookmarkEditorListView::slotClicked(QListViewItem *item)
{
  m_activeItem = static_cast<KFTPBookmarkEditorItem*>(item);

  if ( m_activeItem )
  {
    if (m_activeItem->m_type == BT_CATEGORY) {
      // Category
      m_activeCategory = m_activeItem->m_site;
      m_activeSite = 0L;
    } else {
      // Server
      m_activeSite = m_activeItem->m_site;
      m_activeCategory = m_activeSite->getParentSite();
    }
  }

  emit bookmarkClicked(item);
}

void KFTPBookmarkEditorListView::slotDoubleClicked(QListViewItem *item)
{
  if (!item)
    return;

  setOpen(item, !item->isOpen());
}

KFTPBookmarks::Manager *KFTPBookmarkEditorListView::getBookmarks()
{
  return m_bookmarks;
}

void KFTPBookmarkEditorListView::slotRenameAction()
{
  // Rename a subcategory
  KFTPBookmarkEditorItem *item = static_cast<KFTPBookmarkEditorItem*>(selectedItems().at(0));
  KFTPBookmarks::Site *site = item->m_site;

  bool ok;
  QString newName = KInputDialog::getText(i18n("Category Name"), i18n("Rename category:"), item->text(0), &ok, this);

  if (ok) {
    // Ok, let's rename it
    site->setAttribute("name", newName);
    item->setText(0, newName);
  }
}

void KFTPBookmarkEditorListView::slotSubCatAction()
{
  // Create a subcategory
  KFTPBookmarkEditorItem *item = static_cast<KFTPBookmarkEditorItem*>(selectedItems().at(0));
  KFTPBookmarks::Site *site = item ? item->m_site : m_bookmarks->findCategory("root");

  bool ok;
  QString newName = KInputDialog::getText(i18n("New Category Name"), i18n("New category:"), "", &ok, this);

  if (ok) {
    // Let's create the sub category
    site->addCategory(newName);

    emit categoryRenamed();
    m_bookmarks->emitUpdate();
  }
}

void KFTPBookmarkEditorListView::slotNewAction()
{
  if (!m_activeCategory) {
    // Set the starting category to document root
    m_activeCategory = m_bookmarks->findCategory("root");
  }

  // Create the new node
  KFTPBookmarks::Site *site = m_activeCategory->addSite();
  site->setAttribute("name", i18n("New server"));

  // Create the ListViewItem
  KFTPBookmarkEditorItem *serv = 0L;
  if ( !m_activeItem || ( !m_activeItem->parent() && m_activeItem->m_type == BT_SERVER ) )
    serv = new KFTPBookmarkEditorItem(this, i18n("New Server"));
  else {
    if ( m_activeItem->m_type == BT_CATEGORY ) {
      serv = new KFTPBookmarkEditorItem(m_activeItem, i18n("New Server"));
      m_activeItem->setOpen(true);
    }
    else
      serv = new KFTPBookmarkEditorItem(m_activeItem->parent(), i18n("New Server"));
  }

  serv->m_type = BT_SERVER;
  serv->m_site = site;

  serv->setPixmap(0, loadSmallPixmap("ftp"));

  // Set the new ListViewItem as active
  setSelected(serv, true);
  m_activeItem = serv;

  // Display its properties
  m_activeSite = site;

  emit bookmarkNew(serv, site);
}

void KFTPBookmarkEditorListView::slotDuplicateAction()
{
  // Copy a bookmark
  KFTPBookmarkEditorItem *item = static_cast<KFTPBookmarkEditorItem*>(selectedItems().at(0));

  // Clone the node
  KFTPBookmarks::Site *copy = item->m_site->duplicate();

  // Create the ListViewItem
  KFTPBookmarkEditorItem *serv = 0L;
  if (item->parent())
    serv = new KFTPBookmarkEditorItem(item->parent(), copy->getAttribute("name"));
  else
    serv = new KFTPBookmarkEditorItem(this, copy->getAttribute("name"));

  serv->m_type = BT_SERVER;
  serv->m_site = copy;

  serv->setPixmap(0, loadSmallPixmap("ftp"));

  // Set the new ListViewItem as active
  setSelected(serv, true);
  m_activeItem = serv;

  // Display its properties
  m_activeSite = copy;

  emit bookmarkDuplicated(serv, copy);

  // Update the bookmarks
  m_bookmarks->emitUpdate();
}

void KFTPBookmarkEditorListView::slotDeleteAction()
{
  // Delete a server or a category
  KFTPBookmarkEditorItem *item = static_cast<KFTPBookmarkEditorItem*>(selectedItems().at(0));

  // What do we have here ?
  if (item->m_type == BT_CATEGORY) {
    // Category
    if (KMessageBox::questionYesNo(this, i18n("Are you sure you want to remove this category?")) == KMessageBox::Yes) {
      m_bookmarks->delSite(item->m_site);
    } else {
      return;
    }
  } else {
    // Server
    if (KMessageBox::questionYesNo(this, i18n("Are you sure you want to remove this server?")) == KMessageBox::Yes) {
      m_bookmarks->delSite(item->m_site);
    } else {
      return;
    }
  }

  if ( !m_autoUpdate )
    delete item;

  m_activeItem = 0L;
  emit bookmarkDeleted();
}

void KFTPBookmarkEditorListView::slotContextMenu(QListViewItem *item, const QPoint &pos, int)
{
  slotClicked(item);

  KPopupMenu *context = new KPopupMenu(this);

  if (item)
  {
    context->insertTitle(item->text(0));

    // Server or category
    if (static_cast<KFTPBookmarkEditorItem*>(item)->m_type == BT_CATEGORY) {
      m_newAction->plug(context);
      m_deleteAction->plug(context);
      m_subCatAction->plug(context);
      m_renameAction->plug(context);
    } else {
      if ( m_editMenuItem )
        actionCollection()->action("bookmark_edit2")->plug(context);

      m_newAction->plug(context);
      m_copyAction->plug(context);
      m_deleteAction->plug(context);
    }
  } else {
    // Nothing selected
    m_newAction->plug(context);
    m_subCatAction->plug(context);
  }

  context->exec(pos);

  delete context;
}


#include "kftpbookmarkeditorlistview.moc"
