/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2005 by the KFTPGrabber developers
 * Copyright (C) 2003-2005 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 *
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */


#include <qlabel.h>
#include <qlayout.h>
#include <qspinbox.h>
#include <qlineedit.h>

#include <kdebug.h>
#include <klocale.h>
#include <kpassdlg.h>

#include "qcut.h"
#include "kftpservermanagementlistview.h"
#include "kftpservermanagementdialog.h"
#include "kftpservermanagementdialog.moc"

KFTPServerManagementDialog::KFTPServerManagementDialog()
 : KDialogBase(KDialogBase::Plain, i18n("Site Command Dialog"), KDialogBase::Ok|KDialogBase::Cancel, KDialogBase::Ok)
{
}


KFTPServerManagementDialog::~KFTPServerManagementDialog()
{
}

void KFTPServerManagementDialog::showDialog(KFTPServerManagementListViewItem* item)
{
  // set current item
  m_item = item;
  // get the command from the first column
  m_command = m_item->text(0);

  // check for arguments like <group> in m_command
  checkArguments();
  // create custom dialog
  createDialog();

  if (checkArguments()) {
    // some stuff in the dialog, lets show the dialog
    connect(this, SIGNAL(okClicked()), this, SLOT(slotSendCommand()));
    // show modal dialog
    exec();
  } else {
    // no stuff in the dialog, return the blank site command
    emit command(true, m_command);
  }
}

void KFTPServerManagementDialog::slotSendCommand()
{
  // Remove all characters from < to the end of the line
  // e.g.
  // command before: site stat <value>
  // command after: site stat
  m_command = QCut().cut( "", "<", m_command, 0, 1 );
  m_command = QCut().cut( "", "[", m_command, 0, 1 );

  // username must be entered
  if(m_user == YES) {
    // user didn't enter anything
    if(m_userEdit->text().isEmpty()) {
      emit command(false, i18n("Enter a username or press cancel"));
      return;
    } else {
      m_command += m_userEdit->text();
    }
  }

  // group must be entered
  if(m_group == YES) {
    // user didn't enter anything
    if( m_groupEdit->text().isEmpty()) {
      emit command(false, i18n("Enter a group or press cancel"));
      return;
    } else {
      m_command += m_groupEdit->text();
    }
  }

  // description must be entered
  if(m_description == YES) {
    // user didn't enter anything
    if( m_descriptionEdit->text().isEmpty()) {
      emit command(false, i18n("Enter a description or press cancel"));
      return;
    } else {
      m_command += m_descriptionEdit->text();
    }
  }

  // value in the dialog
  if (m_value) {
    // get the value
    m_command += " " + m_valueSpinBox->text();
  }

  if (m_password) {
    QString pass1 = m_passwordEdit1->password();
    QString pass2 = m_passwordEdit2->password();

    if (pass1 != pass2) {
      emit command(false, i18n("Passwords don't match"));
      return;
    } else {
      m_command += " " + pass1;
    }
  }

  // send the command to KFTPServerManagementListView::slotSendCommand(QString)
  emit command(true, m_command);
}

bool KFTPServerManagementDialog::checkArguments()
{
  bool dialog = false;

  // default
  m_value = NO;
  m_user = NO;
  m_group = NO;
  m_password = NO;
  m_description = NO;

  // user must enter a value
  if (m_command.contains("<value>")) {
    dialog = true;
    if (m_command.contains("[<value>]"))
      m_value = OPTIONAL;
    else
      m_value = YES;
  }

  // user must enter a username
  if (m_command.contains("<user>")) {
    dialog = true;
    if (m_command.contains("[<user>]"))
      m_user = OPTIONAL;
    else
      m_user = YES;
  }

  // user must enter a group
  if (m_command.contains("<group>")) {
    dialog = true;
    if (m_command.contains("[<group>]"))
      m_group = OPTIONAL;
    else
      m_group = YES;
  }

  // user must enter a description
  if (m_command.contains("<description>")) {
    dialog = true;
    if (m_command.contains("[<description>]"))
      m_description = OPTIONAL;
    else
      m_description = YES;
  }

  // user must enter a password
  if (m_command.contains("<password>")) {
    dialog = true;
    m_password = YES;
  }

  return dialog;
}

void KFTPServerManagementDialog::createDialog()
{
  m_layout = new QGridLayout(plainPage(), 1, 1, 11, 6, "dialogLayout");

  // layout integer
  int l = 0;

  // add value stuff to the dialog
  if (m_value) {
    m_valueLabel = new QLabel(i18n("Value:"), plainPage(), "dialogValueLabel");
    m_valueSpinBox = new QSpinBox(plainPage(), "dialogSpinBox");

    // set minimun value
    if (m_item->m_valueMin)
      m_valueSpinBox->setMinValue(m_item->m_valueMin);

    // set maximum value
    if (m_item->m_valueMax)
      m_valueSpinBox->setMaxValue(m_item->m_valueMax);

    // set the label
    if (!m_item->m_valueMinMax.isEmpty())
      if (m_value == YES)
        m_valueLabel->setText(i18n("Value Range (%1)").arg(m_item->m_valueMinMax));
      else
        m_valueLabel->setText(i18n("Value Range (%1) (Optional)").arg(m_item->m_valueMinMax));

    // add to layout
    m_layout->addWidget(m_valueLabel, l, 0);
    m_layout->addWidget(m_valueSpinBox, l, 1);

    ++l;
  }

  if (m_user) {
    m_userLabel = new QLabel(plainPage(), "dialogUserLabel");
    m_userEdit = new QLineEdit(plainPage(), "dialogUserEdit");

    if (m_user == YES)
      m_userLabel->setText(i18n("User name:"));
    else
      m_userLabel->setText(i18n("User name (Optional):"));

    m_layout->addWidget(m_userLabel, l, 0);
    m_layout->addWidget(m_userEdit, l, 1);

    ++l;
  }

  if (m_group) {
    m_groupLabel = new QLabel(plainPage(), "dialogGroupLabel");
    m_groupEdit = new QLineEdit(plainPage(), "dialogGroupEdit");

    if (m_group == YES)
      m_groupLabel->setText(i18n("Group name:"));
    else
      m_groupLabel->setText(i18n("Group name (Optional):"));

    m_layout->addWidget(m_groupLabel, l, 0);
    m_layout->addWidget(m_groupEdit, l, 1);

    ++l;
  }

  if (m_description) {
    m_descriptionLabel = new QLabel(plainPage(), "dialogDescriptionLabel");
    m_descriptionEdit = new QLineEdit(plainPage(), "dialogDescriptionEdit");

    if (m_description == YES)
      m_descriptionLabel->setText(i18n("Description name:"));
    else
      m_descriptionLabel->setText(i18n("Description name (optional):"));

    m_layout->addWidget(m_descriptionLabel, l, 0);
    m_layout->addWidget(m_descriptionEdit, l, 1);

    ++l;
  }

  if (m_password) {
    m_passwordLabel1 = new QLabel(i18n("Password:"), plainPage(), "dialogPasswordLabel1");
    m_passwordLabel2 = new QLabel(i18n("Password retype:"), plainPage(), "dialogPasswordLabel2");
    m_passwordEdit1 = new KPasswordEdit(KPasswordEdit::OneStar, plainPage(), "dialogPasswordEdit1");
    m_passwordEdit2 = new KPasswordEdit(KPasswordEdit::OneStar, plainPage(), "dialogPasswordEdit2");

    m_layout->addWidget(m_passwordLabel1, l, 0);
    m_layout->addWidget(m_passwordEdit1, l++, 1);
    m_layout->addWidget(m_passwordLabel2, l, 0);
    m_layout->addWidget(m_passwordEdit2, l, 1);

    ++l;
  }
}


