/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 * Copyright (C) 2005 Markus Brueffer <markus@brueffer.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "misc.h"

#include "kftplistview.h"
#include "kftpqueueview.h"
#include "kftpqueue.h"
#include "kftpapi.h"
#include "kftpqueueeditor.h"
#include "kftpsearchdialog.h"

#include <kapplication.h>
#include <kfiledialog.h>
#include <kiconloader.h>
#include <kio/job.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kmimetype.h>
#include <kpopupmenu.h>
#include <klistviewsearchline.h>

#include <kdebug.h>

using namespace KFTPGrabberBase;

KFTPQueueViewItem::KFTPQueueViewItem(KFTPQueue::Transfer *transfer, QListView *parent)
  : QObject(parent), QListViewItem(parent)
{
  m_queueTransfer = transfer;

  connect(transfer, SIGNAL(transferUpdate()), this, SLOT(slotUpdateItemRequested()));
  refresh();
}

KFTPQueueViewItem::KFTPQueueViewItem(KFTPQueue::Transfer *transfer, QListViewItem *parent)
  : QObject(), QListViewItem(parent)
{
  m_queueTransfer = transfer;

  connect(transfer, SIGNAL(transferUpdate()), this, SLOT(slotUpdateItemRequested()));
  refresh();
}

void KFTPQueueViewItem::paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int alignment)
{
    QColorGroup _cg( cg );
    QColor c = _cg.text();

    QColor n_color;
    if (m_queueTransfer) {
      switch (m_queueTransfer->getStatus()) {
        case KFTPQueue::Transfer::Running: n_color.setRgb(255, 0, 0); break;
        case KFTPQueue::Transfer::Connecting: n_color.setRgb(0, 0, 255); break;
        case KFTPQueue::Transfer::Locked: n_color.setRgb(0, 150, 0); break;
        default: break;
      }
    }
    
    if (n_color.isValid())
      _cg.setColor(QColorGroup::Text, n_color);

    QListViewItem::paintCell( p, _cg, column, width, alignment );
    _cg.setColor( QColorGroup::Text, c );
}

void KFTPQueueViewItem::refresh()
{
  if (!m_queueTransfer)
    return;

  KFTPQueue::Transfer *transfer = m_queueTransfer;

  // Speed
  QString speed;
  if ((!transfer->isDir() || !isOpen()) && transfer->isRunning())
  {
    speed.sprintf( "%lld KB/s", (transfer->getSpeed() / 1024) );
    
    if (transfer->getSpeed() > 1024*1024)
      speed.sprintf("%lld MB/s", (transfer->getSpeed() / 1024) / 1024);
  
    if (transfer->getSpeed() == 0 && transfer->getTransferType() != KFTPQueue::FXP)
      speed = i18n("stalled");
  }

  // ETA
  QString eta;

  if (transfer->isRunning() && transfer->getSpeed() > 0) {
    filesize_t r_size;
    
    if (transfer->isDir()) {
      r_size = transfer->getActualSize();
    } else {
      r_size = transfer->getSize() - transfer->getCompleted() - transfer->getResumed();
    }
    
    eta = KIO::calculateRemaining(r_size,
                                  0,
                                  transfer->getSpeed()).toString("hh:mm:ss");
  } else {
    eta = "";
  }

  // Set the columns
  setText(0, transfer->getSourceUrl().fileName());
  setText(1, KIO::convertSize(transfer->getActualSize()));
  setText(2, transfer->getSourceUrl().prettyURL());
  setText(3, transfer->getDestUrl().prettyURL());
  setText(5, speed);
  setText(6, eta);
  setText(7, QString::number(transfer->getId()));

  // Progress
  int progress;
  int r_progress;

  if (transfer->getSize() == 0)
    progress = 0;
  else
    progress = transfer->getCompleted()*100/transfer->getSize();
  
  // We can't get progress for FXP transfers
  if (transfer->getTransferType() == KFTPQueue::FXP)
    progress = -1;

  if (transfer->getResumed() == 0)
    r_progress = 0;
  else
    r_progress = transfer->getResumed()*100/transfer->getSize();

  if (m_queueTransfer->getStatus() == KFTPQueue::Transfer::Connecting) {
    // Transfer is not yet connected,
    setText(4, i18n("Connecting..."));
  } else if (progress >= 0) {
    setPixmap(4, createProgressPixmap(progress, r_progress));

    QString progressText;
    progressText.sprintf("%d %%", progress);
    setText(4, progressText);
  } else if (transfer->isRunning() && transfer->getTransferType() == KFTPQueue::FXP) {
    setText(4, i18n("Unknown (FXP)"));
    setText(5, i18n("running"));
  }

  // Transfer type pixmap
  setPixmap(2, loadSmallPixmap(transfer->getSourceUrl().isLocalFile() ? "system" : "network"));
  setPixmap(3, loadSmallPixmap(transfer->getDestUrl().isLocalFile() ? "system" : "network"));

  // Icon
  QString iconText;
  if (transfer->isDir()) {
    iconText = "folder";
  } else {
    KMimeType::Ptr theType = KMimeType::findByURL("/" + transfer->getSourceUrl().path(), 0, false, true);
    iconText = theType->icon(QString::null, false);
  }

  setPixmap(0, loadSmallPixmap(iconText));
  
  // Force repaint
  repaint();
}

void KFTPQueueViewItem::slotUpdateItemRequested()
{
  refresh();
}

void KFTPQueueViewItem::setTransfer(KFTPQueue::Transfer *transfer)
{
  m_queueTransfer = transfer;
}

KFTPQueue::Transfer *KFTPQueueViewItem::getTransfer()
{
  return m_queueTransfer;
}

int KFTPQueueViewItem::compare(QListViewItem *i, int, bool) const
{
  if (!m_queueTransfer) return 1;
  
  KFTPQueue::Transfer *f_transfer = static_cast<KFTPQueueViewItem*>(i)->getTransfer();

  if (f_transfer && f_transfer->isChild() && f_transfer->parent() == m_queueTransfer->parent()) {
    KFTPQueue::TransferDir *f_parent = static_cast<KFTPQueue::TransferDir*>(f_transfer->getTransferParent());

    if (f_parent->getChildrenList().findRef(f_transfer) <
        f_parent->getChildrenList().findRef(m_queueTransfer))
      return 1;
    else
      return -1;
  }
  
  if (FTPQueueManager->getQueueTransferList()->findRef(f_transfer) <
      FTPQueueManager->getQueueTransferList()->findRef(m_queueTransfer))
    return 1;
  else
    return -1;
}

/*
                                                           CLASS KFTPQUEUEVIEW
*/

KActionCollection *KFTPQueueView::actionCollection()
{
  return KFTPAPI::getInstance()->mainWindow()->actionCollection();
}

KFTPQueueView::KFTPQueueView(QWidget *parent, const char *name)
 : QWidget(parent, name)
{
  QVBoxLayout *layout = new QVBoxLayout(this);

  m_toolBar = new KToolBar(this, "queueToolBar");
  m_toolBar->setIconSize(16);
  layout->addWidget(m_toolBar);
  
  m_searchToolBar = new KToolBar(this, "searchToolBar");
  m_searchToolBar->setEnableContextMenu(false);
  m_searchToolBar->setMovingEnabled(false);
  m_searchToolBar->setFullSize(true);

  // Create the erase button
  m_searchToolBar->insertButton(QApplication::reverseLayout() ? "clear_left" :"locationbar_erase", 0, SIGNAL(clicked()), this, SLOT(slotSearchEraseClicked()), true);
  
  // Create the labels
  QLabel *searchLabel = new QLabel(i18n("Filter: "), m_searchToolBar);
  m_searchToolBar->insertWidget(1, 35, searchLabel);
  
  // Create the list view
  m_queue = new KFTPListView(this);
  
  // Create the search field
  m_searchField = new KListViewSearchLine(m_searchToolBar, m_queue);
  
  // Do some more stuff
  m_searchToolBar->setItemAutoSized(1, true);
  m_searchToolBar->setStretchableWidget(m_searchField);
  m_searchToolBar->updateRects(true);
  
  layout->addWidget(m_searchToolBar);
  
  // Create the columns
  m_queue->addColumn(i18n("Name"), 150);
  m_queue->addColumn(i18n("Size"), 75);
  m_queue->addColumn(i18n("Source"), 250);
  m_queue->addColumn(i18n("Destination"), 250);
  m_queue->addColumn(i18n("Progress"), 140);
  m_queue->addColumn(i18n("Speed"), 70);
  m_queue->addColumn(i18n("ETA"), 80);
  
  // Text when there is nothing queued
  m_queue->setEmptyListText(i18n("You don't have any files in the queue."));

  // Multi-select
  m_queue->setSelectionModeExt(KListView::FileManager);
  m_queue->setAllColumnsShowFocus(true);
  m_queue->setRootIsDecorated(true);

  layout->addWidget(m_queue);

  // The signals
  connect(m_queue, SIGNAL(contextMenu(KListView*, QListViewItem*, const QPoint&)), this,
          SLOT(contextMenuRequested(KListView*, QListViewItem*, const QPoint&)));
  connect(m_queue, SIGNAL(selectionChanged()), this, SLOT(updateActions()));

  // Let us be up-to-date
  connect(FTPQueueManager, SIGNAL(queueUpdate()), this, SLOT(slotQueueUpdate()));
  connect(FTPQueueManager, SIGNAL(transferRemoved(long)), this, SLOT(slotTransferRemoved(long)));
  connect(FTPQueueManager, SIGNAL(newTransfer(KFTPQueue::Transfer*)), this, SLOT(slotTransferAdded(KFTPQueue::Transfer*)));
  connect(FTPQueueManager, SIGNAL(queueUpdate()), this, SLOT(updateActions()));
  
  // Load the listview layout
  loadLayout();
  
  // Create the context menu actions
  initActions();
  initToolBar();
  updateActions();
}

KFTPQueueView::~KFTPQueueView()
{
}

void KFTPQueueView::saveLayout()
{
  m_queue->saveLayout(kapp->config(), "queueViewLayout");
}

void KFTPQueueView::loadLayout()
{
  m_queue->restoreLayout(kapp->config(), "queueViewLayout");
}

void KFTPQueueView::initToolBar()
{
  m_loadAction->plug(m_toolBar);
  m_saveAction->plug(m_toolBar);
  m_toolBar->insertSeparator();
  m_startAction->plug(m_toolBar);
  m_pauseAction->plug(m_toolBar);
  m_stopAction->plug(m_toolBar);
  m_toolBar->insertSeparator();
  m_addAction->plug(m_toolBar);
  m_removeAction->plug(m_toolBar);
  m_searchAction->plug(m_toolBar);
}

void KFTPQueueView::initActions()
{
  // Create all the actions
  m_launchAction = new KAction(i18n("&Start Transfer"), "launch", KShortcut(), this, SLOT(slotLaunch()), actionCollection(), "launch");
  m_abortAction = new KAction(i18n("&Abort Transfer"), KShortcut(), this, SLOT(slotAbort()), actionCollection(), "abort");
  m_removeAction = new KAction(i18n("&Remove"), "editdelete", KShortcut(), this, SLOT(slotRemove()), actionCollection(), "remove");
  m_removeAllAction = new KAction(i18n("Remove &All"), KShortcut(), this, SLOT(slotRemoveAll()), actionCollection(), "removeAll");
  m_moveUpAction = new KAction(i18n("Move &Up"), "up", KShortcut(), this, SLOT(slotMoveUp()), actionCollection(), "moveUp");
  m_moveDownAction = new KAction(i18n("Move &Down"), "down", KShortcut("del"), this, SLOT(slotMoveDown()), actionCollection(), "moveDown");
  m_editAction = new KAction(i18n("&Change Transfer Info"), KShortcut(), this, SLOT(slotEdit()), actionCollection(), "changeTransfer");

  // Create the toolbar actions
  m_loadAction = new KAction(i18n("&Load Queue From File"), "fileopen", KShortcut(), this, SLOT(slotLoad()), actionCollection(), "load");
  m_saveAction = new KAction(i18n("&Save Queue to File"), "filesaveas", KShortcut(), this, SLOT(slotSave()), actionCollection(), "save");
  m_startAction = new KAction(i18n("S&tart"), "player_play", KShortcut(), this, SLOT(slotStart()), actionCollection(), "start");
  m_pauseAction = new KAction(i18n("&Pause"), "player_pause", KShortcut(), this, SLOT(slotPause()), actionCollection(), "pause");
  m_stopAction = new KAction(i18n("St&op"), "player_stop", KShortcut(), this, SLOT(slotStop()), actionCollection(), "stop");
  m_addAction = new KAction(i18n("&Add Transfer..."), "filenew", KShortcut(), this, SLOT(slotAdd()), actionCollection(), "add");
  m_searchAction = new KAction(i18n("&Search && Replace..."), "find", KShortcut(), this, SLOT(slotSearch()), actionCollection(), "search");
    
  m_saveAction->setEnabled( false );
  m_startAction->setEnabled(false);
  m_pauseAction->setEnabled(false);
  m_stopAction->setEnabled(false);
  m_addAction->setEnabled(false);
  m_removeAction->setEnabled(false);
  m_searchAction->setEnabled( false );
}

void KFTPQueueView::updateActions()
{
  m_startAction->setEnabled(!FTPQueueManager->queueProcRunning() && FTPQueueManager->getQueueTransferList()->count() > 0 && FTPQueueManager->getNumRunning() == 0);
  m_stopAction->setEnabled(FTPQueueManager->queueProcRunning());
  m_removeAllAction->setEnabled(!FTPQueueManager->queueProcRunning());
  
  QPtrList<QListViewItem> selection = m_queue->selectedItems();
  KFTPQueueViewItem *firstItem = static_cast<KFTPQueueViewItem*>(selection.first());
  
  m_removeAction->setEnabled((bool)firstItem);
  
  if (!firstItem)
    return;
    
  bool parentRunning = false;
  if (firstItem->getTransfer()->isChild())
    parentRunning = firstItem->getTransfer()->getTransferParent()->isRunning();
  
  m_launchAction->setEnabled(!firstItem->getTransfer()->isRunning() && !FTPQueueManager->queueProcRunning() && !firstItem->getTransfer()->isLocked());
  m_abortAction->setEnabled(firstItem->getTransfer()->isRunning() && !FTPQueueManager->queueProcRunning());
  m_removeAction->setEnabled(!firstItem->getTransfer()->isRunning() && !FTPQueueManager->queueProcRunning() && !firstItem->getTransfer()->isLocked());
  m_editAction->setEnabled(!firstItem->getTransfer()->isRunning() && !firstItem->getTransfer()->isChild() && !firstItem->getTransfer()->isLocked());
  
  // Only allow moving of multi selections if they have the same parent
  bool allowMove = true;
  for (QListViewItem *i = selection.first(); i; i = selection.next()) {
    if ( i->parent() != static_cast<QListViewItem*>(firstItem)->parent() ) {
      allowMove = false;
      break;
    }
  }
  
  m_moveUpAction->setEnabled(allowMove && FTPQueueManager->canBeMovedUp(firstItem->getTransfer()->getId()) && !firstItem->getTransfer()->isLocked());
  m_moveDownAction->setEnabled(allowMove && FTPQueueManager->canBeMovedDown(static_cast<KFTPQueueViewItem*>(selection.last())->getTransfer()->getId()) && !firstItem->getTransfer()->isLocked());
}

void KFTPQueueView::slotTransferAdded(KFTPQueue::Transfer *transfer)
{
  if (transfer->isChild()) {
    // This transfer should be inserted under some other transfer
    QListViewItem *parent = m_queue->findItem(QString::number(transfer->getTransferParent()->getId()), 7);
    
    if (parent)
      new KFTPQueueViewItem(transfer, parent);
  }
  else
    new KFTPQueueViewItem(transfer, m_queue);

  // Update actions
  m_saveAction->setEnabled( true );
  m_removeAllAction->setEnabled( true );
  m_searchAction->setEnabled( true );
}

void KFTPQueueView::slotTransferRemoved(long id)
{
  // Delete the transfer
  QListViewItem *item = m_queue->findItem(QString::number(id), 7);
  
  if (item)
    delete item;

  // Update actions
  bool empty = (m_queue->childCount() == 0);

  m_saveAction->setEnabled( !empty );
  if (empty) m_removeAction->setEnabled( false );
  m_removeAllAction->setEnabled( !empty );
  m_searchAction->setEnabled( !empty );
}

void KFTPQueueView::slotQueueUpdate()
{
  /* Do a resort */
  m_queue->sort();
}

void KFTPQueueView::contextMenuRequested(KListView*, QListViewItem* item, const QPoint& p)
{
  if (!item)
    return;

  KPopupMenu *contextMenu = new KPopupMenu(m_queue);
  
  // Populate context menu
  contextMenu->insertTitle(item->text(0) + ((m_queue->selectedItems().count() > 1) ? "..." : "" ));
  m_launchAction->plug(contextMenu);
  m_abortAction->plug(contextMenu);
  contextMenu->insertSeparator();
  m_removeAction->plug(contextMenu);
  m_removeAllAction->plug(contextMenu);
  contextMenu->insertSeparator();
  m_moveUpAction->plug(contextMenu);
  m_moveDownAction->plug(contextMenu);
  contextMenu->insertSeparator();
  m_editAction->plug(contextMenu);
  
  // Update the actions
  updateActions();
  
  // Show the context menu
  contextMenu->exec(p);
}

void KFTPQueueView::slotLaunch()
{
  static_cast<KFTPQueueViewItem*>(m_queue->selectedItems().first())->getTransfer()->execute();
}

void KFTPQueueView::slotAbort()
{
  static_cast<KFTPQueueViewItem*>(m_queue->selectedItems().first())->getTransfer()->abort();
}

void KFTPQueueView::slotRemove()
{
  if (KMessageBox::questionYesNo(this, i18n("Are you sure you want to remove queued file(s)?")) == KMessageBox::Yes) {
    FTPQueueManager->setEmitUpdate(false);
   
    QPtrList<QListViewItem> selection = m_queue->selectedItems();
    for (QListViewItem *item = selection.first(); item; item = selection.next()) {
      if (item && static_cast<KFTPQueueViewItem*>(item)->getTransfer())
        FTPQueueManager->removeTransfer(static_cast<KFTPQueueViewItem*>(item)->getTransfer());
    }
    
    FTPQueueManager->setEmitUpdate(true);
    FTPQueueManager->doEmitUpdate();
  }
}

void KFTPQueueView::slotRemoveAll()
{
  if (KMessageBox::questionYesNo(this, i18n("Are you sure you want to remove ALL queued files?")) == KMessageBox::Yes) {
    FTPQueueManager->clearQueue();
  }
}

void KFTPQueueView::slotMoveUp()
{
  QPtrList<QListViewItem> selection = m_queue->selectedItems();
  
  for (QListViewItem *item = selection.first(); item; item = selection.next())
  {
    // Move the listview item
    // XXX: Is there a better/faster way to do this?
    QListViewItem *newItemPos = 0, *prevItem;
    
    if (item->parent())
      prevItem = item->parent()->firstChild();
    else
      prevItem = m_queue->firstChild();
      
    for (QListViewItem *i = prevItem; i; i = i->nextSibling())
    {
      if (i == item)
      {
        m_queue->moveItem(item, item->parent(), newItemPos);
        break;
      }
      
      newItemPos = prevItem;
    }
    
    // Move the transfer
    FTPQueueManager->moveTransferUp(static_cast<KFTPQueueViewItem*>(item)->getTransfer()->getId());
  }
}

void KFTPQueueView::slotMoveDown()
{
  QPtrList<QListViewItem> selection = m_queue->selectedItems();
  for (QListViewItem *item = selection.last(); item; item = selection.prev())
  {
    // Move the listview item
    QListViewItem *below = item->nextSibling();
    m_queue->moveItem(item, item->parent(), below);
    
    // Move the transfer
    FTPQueueManager->moveTransferDown(static_cast<KFTPQueueViewItem*>(item)->getTransfer()->getId());
  }
}

void KFTPQueueView::slotEdit()
{
  KFTPQueueEditor *editor = new KFTPQueueEditor(this);
  
  KFTPQueueViewItem* item = static_cast<KFTPQueueViewItem*>(m_queue->selectedItems().first());
  editor->setData(item->getTransfer());
  
  // Show the queue editor
  if (editor->exec() == QDialog::Accepted) {
    editor->saveData();
    
    FTPQueueManager->doEmitUpdate();
    item->refresh();
  }
}

void KFTPQueueView::slotSearch()
{
  KFTPSearchDialog *dialog = new KFTPSearchDialog();
  
  dialog->exec();
  delete dialog;
}

void KFTPQueueView::slotLoad()
{
  if (m_queue->childCount() && KMessageBox::warningContinueCancel(0L, i18n("Loading a new queue will overwrite the existing one! Are you sure you want to continue?"), i18n("Load Queue")) == KMessageBox::Cancel)
    return;

  QString loadPath = KFileDialog::getOpenFileName();
  
  if (!loadPath.isEmpty()) {
    FTPQueueManager->getConverter()->importQueue(loadPath);
  }
}

void KFTPQueueView::slotSave()
{
  QString savePath = KFileDialog::getSaveFileName();
  
  if (!savePath.isEmpty()) {
    FTPQueueManager->getConverter()->exportQueue(savePath);
  }
}

void KFTPQueueView::slotStart()
{
  // Begin queue processing
  FTPQueueManager->startQueueTransfer();
}

void KFTPQueueView::slotPause()
{
}

void KFTPQueueView::slotStop()
{
  // Abort queue processing
  FTPQueueManager->abortQueueTransfer();
}

void KFTPQueueView::slotAdd()
{
}

void KFTPQueueView::slotSearchEraseClicked()
{
  m_searchField->clear();
}


#include "kftpqueueview.moc"
